/* iksemel (XML parser for Jabber)
** Copyright (C) 2000-2003 Gurer Ozen <madcat@e-kolay.net>
** This code is free software; you can redistribute it and/or
** modify it under the terms of GNU Lesser General Public License.
*/

/* This file has been modified by
** Siemens Home and Office Communication Devices GmbH & Co. KG
** (SHC) 
** http://gigaset.siemens.com
** 2006-02-21: some casts
** 2006-11-02: static variables "*(*my_malloc_func)(size_t size)" and "(*my_free_func)(void *ptr)" changed to extern defined
** 2007-11-19: some code optimisations
*/

#ifndef SHC_VERSION
#include "common.h"
#include "iksemel.h"
#endif

/*****  malloc wrapper  *****/
#ifndef SHC_VERSION
static void *(*my_malloc_func)(size_t size);
static void (*my_free_func)(void *ptr);
#else
extern void *(*my_malloc_func)(size_t size);
extern void (*my_free_func)(void *ptr);
#endif

void *
iks_malloc (size_t size)
{
#ifndef SHC_VERSION
  if (my_malloc_func)
#endif
		return my_malloc_func (size);
#ifndef SHC_VERSION
	else
		return malloc (size);
#endif
}

void
iks_free (void *ptr)
{
#ifndef SHC_VERSION
  if (my_free_func)
#endif
		my_free_func (ptr);
#ifndef SHC_VERSION
	else
		free (ptr);
#endif
}

#ifndef SHC_VERSION
void
iks_set_mem_funcs (void *(*malloc_func)(size_t size), void (*free_func)(void *ptr))
{
	my_malloc_func = malloc_func;
	my_free_func = free_func;
}
#endif

/*****  NULL-safe Functions  *****/
#ifndef SHC_VERSION
char *
iks_strdup (const char *src)
{  
  if (src) return strdup(src);
  return NULL;
}
#endif

#ifndef SHC_VERSION
char *
iks_strcat (char *dest, const char *src)
{
	size_t len;

	if (!src) return dest;

	len = strlen (src);
	memcpy (dest, src, len);
	dest[len] = '\0';
	return dest + len;
}
#endif

int
iks_strcmp (const char *a, const char *b)
{
	if (!a || !b) return -1;
	return strcmp (a, b);
}

int
iks_strcasecmp (const char *a, const char *b)
{
	if (!a || !b) return -1;
  return strcasecmp (a, b);
}

#ifndef SHC_VERSION
int
iks_strncmp (const char *a, const char *b, size_t n)
{
	if (!a || !b) return -1;
	return strncmp (a, b, n);
}
#endif

#ifndef SHC_VERSION
int
iks_strncasecmp (const char *a, const char *b, size_t n)
{
	if (!a || !b) return -1;
  return strncasecmp (a, b, n);
}
#endif

#ifndef SHC_VERSION
size_t
iks_strlen (const char *src)
{
	if (!src) return 0;
	return strlen (src);
}
#endif

/*****  XML Escaping  *****/

#ifndef SHC_VERSION
char *
iks_escape (ikstack *s, char *src, size_t len)
{
	char *ret;
  int i, j, nlen;

	if (!src || !s) return NULL;
	if (len == -1) len = strlen (src);

	nlen = len;
	for (i=0; i<len; i++) {
		switch (src[i]) {
		case '&': nlen += 4; break;
		case '<': nlen += 3; break;
		case '>': nlen += 3; break;
		case '\'': nlen += 5; break;
		case '"': nlen += 5; break;
		}
	}
	if (len == nlen) return src;
 
  ret = iks_stack_alloc (s, nlen + 1);
	if (!ret) return NULL;

	for (i=j=0; i<len; i++) {
		switch (src[i]) {
		case '&': memcpy (&ret[j], "&amp;", 5); j += 5; break;
		case '\'': memcpy (&ret[j], "&apos;", 6); j += 6; break;
		case '"': memcpy (&ret[j], "&quot;", 6); j += 6; break;
		case '<': memcpy (&ret[j], "&lt;", 4); j += 4; break;
		case '>': memcpy (&ret[j], "&gt;", 4); j += 4; break;
		default: ret[j++] = src[i];
		}
	}
	ret[j] = '\0';

	return ret;
}
#endif

#ifndef SHC_VERSION
char *
iks_unescape (ikstack *s, char *src, size_t len)
{
  int i,j;
	char *ret;

	if (!s || !src) return NULL;
	if (!strchr (src, '&')) return src;
	if (len == -1) len = strlen (src);
  
  ret = iks_stack_alloc (s, len + 1);
	if (!ret) return NULL;

	for (i=j=0; i<len; i++) {
		if (src[i] == '&') {
			i++;
			if (strncmp (&src[i], "amp;", 4) == 0) {
				ret[j] = '&';
				i += 3;
			} else if (strncmp (&src[i], "quot;", 5) == 0) {
				ret[j] = '"';
				i += 4;
			} else if (strncmp (&src[i], "apos;", 5) == 0) {
				ret[j] = '\'';
				i += 4;
			} else if (strncmp (&src[i], "lt;", 3) == 0) {
				ret[j] = '<';
				i += 2;
			} else if (strncmp (&src[i], "gt;", 3) == 0) {
				ret[j] = '>';
				i += 2;
			} else {
				ret[j] = src[--i];
			}
		} else {
			ret[j] = src[i];
		}
		j++;
	}
	ret[j] = '\0';

	return ret;
}
#endif
